// Scorer for TAC KBP 2010 slot-filling task
// author:  Ralph Grishman

// version 0.8
// January 15, 2010

import java.io.*;
import java.util.*;

public class SlotScorer {

    // true to print out judgement for each line of response
    static boolean trace = false;
    // true to replicate Paul's scorer (no credit for redundant answers)
    static boolean official2009 = true;

    // tables built from judgement file
    static Map<String, Integer> judgement = new HashMap<String, Integer> ();
    static Map<String, Integer> equivalenceClass = new HashMap<String, Integer> ();
    static Map<String, Boolean> query_has_answer = new HashMap<String, Boolean> ();
    static Map<String, Set<Integer>> query_eclasses = new HashMap<String, Set<Integer>> ();

    static Map<String, List<String>> response = new HashMap <String, List<String>> ();

    // codes in judgement file
    static final int WRONG = -1;
    static final int CORRECT = 1;
    static final int REDUNDANT = 2;
    static final int INEXACT = 3;

    /**
     *  SlotScorer <response file>
     *  scores response file using files
     *     alljudgements - list of assessments
     *     sv-slots - list of single-valued slots
     *     l-slots - list of list-valued slots
     */

    public static void main (String[] args) throws IOException {

	if (args.length != 1) {
	    System.out.println ("SlotScorer must be invoked with 1 argumennt:  response file");
	    System.exit(1);
	}
	String responseFile = args[0];

	// ----------- read in slot judgements ------------

	String keyFile = "alljudgements";
	BufferedReader keyReader = null;
	try {
	    keyReader = new BufferedReader (new FileReader(keyFile));
	} catch (FileNotFoundException e) {
	    System.out.println ("Unable to open judgement file " + keyFile);
	    System.exit (1);
	}
	String line;
	while ((line = keyReader.readLine()) != null) {
	    String[] fields = line.split(" ", 8);
	    if (fields.length != 8) {
		System.out.println ("Invalid line in judgement file:");
		System.out.println (line);
		continue;
	    }
	    String query_id = fields[1].replace(',', '/');
	    String doc_id = fields[2];
	    String answerString = fields[7];
	    int jment = 0;
	    try {
		jment = Integer.parseInt(fields[4]);
	    } catch (NumberFormatException e) {
		System.out.println ("Invalid line in judgement file -- invalid judgement:");
		System.out.println (line);
		continue;
	    }
	    int eclass = 0;
	    try {
		eclass = Integer.parseInt(fields[5]);
	    } catch (NumberFormatException e) {
		System.out.println ("Invalid line in judgement file -- invalid equivalence class:");
		System.out.println (line);
		continue;
	    }
	    String key = query_id + ":" + doc_id + ":" + answerString;
	    judgement.put(key, jment);
	    equivalenceClass.put(key, eclass);
	    if (jment == CORRECT) {
		query_has_answer.put(query_id, true);
		if (query_eclasses.get(query_id) == null)
		    query_eclasses.put(query_id, new HashSet<Integer>());
		query_eclasses.get(query_id).add(eclass);
	    }
	}
	System.out.println ("Read " + judgement.size() + " judgements.");

	// --------- read in system responses -------------

	BufferedReader responseReader = null;
	try {
	    responseReader = new BufferedReader (new FileReader(responseFile));
	} catch (FileNotFoundException e) {
	    System.out.println ("Unable to open response file " + responseFile);
	    System.exit (1);
	}
	// String line;
	while ((line = responseReader.readLine()) != null) {
	    String[] fields = line.trim().split("\\s+", 6);
	    if (!((fields.length == 4) | (fields.length == 6))) {
		System.out.println ("Invalid line in response file:");
		System.out.println (line);
		continue;
	    }
	    String entity = fields[0];
	    String slot = fields[1];
	    String query_id = entity + ":" + slot;
	    String doc_id = fields[3];
	    String answer_string = "";
	    if (fields.length == 6)
		answer_string = ":" + fields[5];
	    if (response.get(query_id) == null)
		response.put(query_id, new ArrayList<String>());
	    response.get(query_id).add(doc_id + answer_string);
	}
	System.out.println ("Read responses for " + response.size() + " slots.");

	// -------------- read list of slots ----------

	List<String> svSlots = readLines("sv-slots");
	List<String> lSlots = readLines("l-slots");

	// ------------- score responses ------------
	//          for single-valued slots

	// correct single-valued slots, including NILs
	int num_right = 0;
	int num_wrong = 0;
	int num_inexact = 0;
	int num_redundant = 0;
	int num_nils = 0;
	// number of non-NIL responses
	int num_responses = 0;
	// number of correct non-NIL responses
	int num_good_responses = 0;
	// number of correct answers (correct single-value answers + list-value equivalence classes)
	int num_answers = 0;;
	String symbol = "?";

	for (String query : svSlots) {
	    if (query_has_answer.get(query) != null)
		num_answers++;
	    List<String> responseList = response.get(query);
	    if (responseList == null) {
		System.out.println ("No system response for slot " + query);
		continue;
	    }
	    String responseString = responseList.get(0);
	    String fields[] = responseString.split(":",2);
	    String doc_id = fields[0];
	    String answer_string = "";
	    if (fields.length == 2)
		answer_string = fields[1];
	    if (doc_id.equals("NIL")) {
		num_nils++;
		if (query_has_answer.get(query) != null) {
		    num_wrong++;
		    symbol = "W";
		} else {
		    num_right++;
		    symbol = "C";
		}
	    } else {
		num_responses++;
		Integer J = judgement.get(query + ":" + doc_id + ":" + answer_string);
		if (J == null) {
		    System.out.println ("No judgement for " +
			query + ":" + doc_id + " " + answer_string);
		    continue;
		}
		int j = J;
		switch (j) {
		    case WRONG:
			num_wrong++;
			symbol = "W";
			break;
		    case REDUNDANT:
			if (official2009) {
			    num_redundant++;
			    symbol = "R";
			    break;
			}
		    case CORRECT:
			num_right++;
			num_good_responses++;
			symbol = "C";
			break;
		    case INEXACT:
			num_inexact++;
			symbol = "X";
			break;
		    default:
			System.out.println ("Invalid judgement " + j);
		}
	    }
	    if (trace)
		System.out.println (symbol + " " + query + " " + responseString);
	}
	int num_queries = svSlots.size();
	float accuracy = (float) num_right / (float) num_queries;
	System.out.println ("\n2009 scores:");
	System.out.println ("Statistics over " + num_queries + " single valued slots");
	System.out.println ("\tNumber correct (C): " + num_right);
	System.out.println ("\tNumber wrong (W): " + num_wrong);
	if (official2009)
	    System.out.println ("\tNumber redundant (R): " + num_redundant);
	System.out.println ("\tNumber inexact (X): " + num_inexact);
	System.out.println ("\tAccuracy: " + accuracy);

	if (num_responses > 0)
	    System.out.println ("\tPrecision of non-NIL answers: " +
		num_good_responses + "/" + num_responses + " = " +
		(float) num_good_responses / (float) num_responses);
	if (num_answers > 0)
	    System.out.println ("\tRecall of non-NIL answers: " +
		num_good_responses + "/" + num_answers + " = " +
		(float) num_good_responses / (float) num_answers);

	// ------------- score responses ------------
	//           for list-valued slots
 
	float totalF = 0;
	for (String query : lSlots) {
	    List<String> responseList = response.get(query);
	    if (responseList == null) {
		System.out.println ("No system response for slot " + query);
		continue;
	    }
	    Set<Integer> distincts = new HashSet<Integer>();
	    int num_responses_to_query = responseList.size();;
	    for (String responseString : responseList) {
		String fields[] = responseString.split(":",2);
		String doc_id = fields[0];
		String answer_string = "";
		if (fields.length == 2)
		    answer_string = fields[1];
		if (doc_id.equals("NIL")) {
		    if (num_responses_to_query > 1)
			System.out.println ("More than one response, including NIL, for " + query);
		    num_responses_to_query = 0;
		    if (query_has_answer.get(query) != null) {
			symbol = "W";
		    } else {
			symbol = "C";
		    }
		} else {
		    String key = query + ":" + doc_id + ":" + answer_string;
		    Integer J = judgement.get(key);
		    if (J == null) {
			System.out.println ("No judgement for " + key);
			continue;
		    }
		    int j = J;
		    switch (j) {
		    case WRONG:
			num_wrong++;
			symbol = "W";
			break;
		    case REDUNDANT:
			if (official2009) {
			    num_redundant++;
			    symbol = "R";
			    break;
			}
		    case CORRECT:
			Integer E = equivalenceClass.get(key);
			if (distincts.contains(E)) {
			    num_redundant++;
			    symbol = "R";
			} else {
			    num_right++;
			    symbol = "C";
			    distincts.add(E);
			}
			break;
		    case INEXACT:
			num_inexact++;
			symbol = "X";
			break;
		    default:
			System.out.println ("Invalid judgement " + j);
		    }
		}
	    if (trace)
		System.out.println (symbol + " " + query + " " + responseString);
	    }
	    int correct = distincts.size();
	    int num_answers_to_query = 0;
	    if (query_eclasses.get(query) != null)
		num_answers_to_query = query_eclasses.get(query).size();
	    float recall;
	    float precision;
	    if (num_answers_to_query == 0) {
		if (num_responses_to_query == 0) {
		    recall = 1;
		    precision = 1;
		} else {
		    recall = 0;
		    precision = 0;
		}
	    } else /* num_answers_to_query > 0 */ {
		if (num_responses_to_query == 0) {
		    recall = 0;
		    precision = 0;
		} else { 
	    	    recall = ((float) correct) / num_answers_to_query;
	    	    precision = ((float) correct) / num_responses_to_query;
		}
	    }
	    float queryF;
	    if (recall == 0 && precision == 0)
		queryF = 0;
	    else
	        queryF = (5 * recall * precision) / (4 * recall + precision);
	    totalF += queryF;

	    num_responses += num_responses_to_query;
	    num_good_responses += correct;
	    num_answers += num_answers_to_query;
	}
	// compute and report macro average F and overall 2009 score
	float avgF = totalF / lSlots.size();
	System.out.println ("Average F over " + lSlots.size() + " list-valued slots: " + avgF);
	float SFscore = (accuracy + avgF) / 2;
	System.out.println ("SF-value score: " + SFscore);

	float recall = ((float) num_good_responses) / num_answers;
	float precision = ((float) num_good_responses) / num_responses;
	float F = (2 * recall * precision) / (recall + precision);
	System.out.println ("\nProposed 2010 scores:");
	System.out.println ("\tRecall: " + num_good_responses + " / " + num_answers + " = " + recall);
	System.out.println ("\tPrecision: " + num_good_responses + " / " + num_responses + " = " + precision);
	System.out.println ("\tF1: " + F);
    }

    /**
     *  reads a series of lines from 'fileName' and returns them as a list of Strings
     */

    static List<String> readLines (String fileName) {
	BufferedReader reader = null;
	List<String> lines = new ArrayList<String>();
	try {
	    reader = new BufferedReader (new FileReader(fileName));
	} catch (FileNotFoundException e) {
	    System.out.println ("Unable to open file " + fileName);
	    System.exit (1);
	}
	String line;
	try {
	    while ((line = reader.readLine()) != null) {
		lines.add(line.trim());
	    }
	} catch (IOException e) {
	    System.out.println ("Error readng from file " + fileName);
	    System.exit (1);
	}
	System.out.println ("Read " + lines.size() + " lines from " + fileName);
	return lines;
    }
}
